package de.fitko.fitconnect.client.di;

import com.nimbusds.jose.jwk.RSAKey;
import de.fitko.fitconnect.api.domain.auth.OAuthToken;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import de.fitko.fitconnect.api.domain.model.metadata.data.Data;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionSubmit;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;
import de.fitko.fitconnect.api.exceptions.ClientNotAuthenticatedException;
import de.fitko.fitconnect.api.services.Sender;

import java.io.File;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.UUID;
import java.util.stream.Collectors;

/**
 * A fluent client for announcing and handing in a {@link SubmissionSubmit}
 */
public class FluentSenderClient extends Client {

    private FluentSenderClient() {
    }

    public static WithDestination build() {
        final Sender sender = getService(Sender.class);
        final String clientId = getProperty("clientId");
        final String secret = getProperty("clientSecret");
        return new ClientBuilder(sender, clientId, secret);
    }

    public interface WithDestination {

        /**
         * Configures the client for the given destination and load the public key
         *
         * @param destinationId the clients destination
         * @return the upload step for attachments
         */
        WithAttachments withDestination(UUID destinationId);

        SubmissionForPickup send(UUID destinationId, List<File> attachments, String data);

        SubmissionForPickup send(UUID destinationId, List<File> attachments, byte[] data);

        SubmissionForPickup send(UUID destinationId, List<File> attachments);
    }

    public interface WithAttachments {

        /**
         * Sends the submission with a list of attachments
         *
         * @param attachments that are sent with the submission
         * @return the step where additional data can be added to the submission
         */
        WithData withAttachments(List<File> attachments);
    }

    public interface WithData {

        Submit withData(String data);

        Submit withData(byte[] data);

        SubmissionForPickup submit();
    }

    public interface Submit {
        SubmissionForPickup submit();
    }

    public static class ClientBuilder implements WithDestination, WithAttachments, WithData, Submit {

        private final Sender sender;
        private final String clientId;
        private final String secret;

        private UUID destinationId;
        private Data data;
        private List<Attachment> attachments;
        private RSAKey encryptionKey;
        private UUID submissionId;
        private OAuthToken token;

        public ClientBuilder(final Sender sender, final String clientId, final String secret) {
            this.sender = sender;
            this.clientId = clientId;
            this.secret = secret;
            authenticate();
        }

        @Override
        public WithAttachments withDestination(final UUID destinationId) {
            final RSAKey encryptionKey = sender.getEncryptionKeyForDestination(destinationId);
            final ValidationResult validationResult = sender.validatePublicKey(encryptionKey);
            if (validationResult.hasError()) {
                throw new RuntimeException("Public encryption key is not valid", validationResult.getError());
            }
            this.encryptionKey = encryptionKey;
            return this;
        }

        @Override
        public SubmissionForPickup send(final UUID destinationId, final List<File> attachments, final String data) {
            return this.withDestination(destinationId)
                    .withAttachments(attachments)
                    .withData(data)
                    .submit();
        }

        @Override
        public SubmissionForPickup send(final UUID destinationId, final List<File> attachments, final byte[] data) {
            return this.withDestination(destinationId)
                    .withAttachments(attachments)
                    .withData(data)
                    .submit();
        }

        @Override
        public SubmissionForPickup send(final UUID destinationId, final List<File> attachments) {
            return this.withDestination(destinationId)
                    .withAttachments(attachments)
                    .submit();
        }

        @Override
        public Submit withData(final String data) {
            return withData(data.getBytes(StandardCharsets.UTF_8));
        }

        @Override
        public Submit withData(final byte[] data) {
            final var unencryptedData = Data.builder().build();
            final Data encryptedData = sender.encryptSubmissionData(encryptionKey, unencryptedData);
            this.data = encryptedData;
            return this;
        }

        @Override
        public WithData withAttachments(final List<File> attachmentFiles) {
            final List<Attachment> attachments = readFilesToAttachments(attachmentFiles);
            final SubmissionForPickup submission = createSubmission(destinationId, attachments);
            this.submissionId = submission.getSubmissionId();
            final List<Attachment> encryptedAttachments = encryptAttachments(encryptionKey, attachments);
            sender.uploadAttachments(submissionId, encryptedAttachments);
            return this;
        }

        @Override
        public SubmissionForPickup submit() {
            return sender.sendSubmission(submissionId, getMetadata());
        }

        private Metadata getMetadata() {
            return this.data == null ? sender.createMetadata(attachments) : sender.createMetadata(data, attachments);
        }

        private List<Attachment> encryptAttachments(final RSAKey encryptionKey, final List<Attachment> attachments) {
            return attachments.stream().map(attachment -> encryptAttachment(encryptionKey, attachment)).collect(Collectors.toList());
        }

        private Attachment encryptAttachment(final RSAKey encryptionKey, final Attachment attachment) {
            return sender.encryptAttachment(encryptionKey, attachment);
        }

        private SubmissionForPickup createSubmission(final UUID destinationId, final List<Attachment> attachments) {
            final SubmissionSubmit request = SubmissionSubmit.builder()
                    .destinationId(destinationId)
                    .announcedAttachments(asListOfAttachmentsIds(attachments))
                    .build();

            return sender.createSubmission(request);
        }

        private List<UUID> asListOfAttachmentsIds(final List<Attachment> attachments) {
            return attachments.stream().map(attachment -> attachment.getAttachmentId()).collect(Collectors.toList());
        }

        private void authenticate() {
            final OAuthToken oAuthToken = sender.retrieveOAuthToken(clientId, secret);
            if (oAuthToken == null) {
                throw new ClientNotAuthenticatedException("Could not authenticate client, please authenticate first");
            }
        }

        private List<Attachment> readFilesToAttachments(final List<File> attachmentFiles) {
            return attachmentFiles.stream()
                    .map(this::readFileToAttachment)
                    .collect(Collectors.toList());
        }

        private Attachment readFileToAttachment(final File file) {
            return sender.createAttachment(file);
        }
    }
}